using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Xml;

namespace System.Xml.Expression
{
  public class Xexpression
  {
    public static int debug_level = 0;

    public struct Symbol : IEquatable<Symbol>
    {
      private static NameTable nt = new NameTable ();

      internal string name;

      public Symbol (string str)
	{
	  name = nt.Add (str);
	}

      public static implicit operator Symbol (string str)
      {
	return new Symbol (str);
      }

      public static implicit operator string (Symbol name)
      {
	return name.name;
      }

      public static bool operator== (Symbol n1, Symbol n2)
	{
	  return (object) n1.name == (object) n2.name;
	}

      public static bool operator!= (Symbol n1, Symbol n2)
	{
	  return (object) n1.name != (object) n2.name;
	}

      public static bool operator== (Symbol n1, string n2)
	{
	  return (object) n1.name == (object) n2;
	}

      public static bool operator!= (Symbol n1, string n2)
	{
	  return (object) n1.name != (object) n2;
	}

      public static bool operator== (string n1, Symbol n2)
	{
	  return (object) n1 == (object) n2.name;
	}

      public static bool operator!= (string n1, Symbol n2)
	{
	  return (object) n1 != (object) n2.name;
	}

      public bool Equals (Symbol name)
      {
	return Object.ReferenceEquals (this.name, name.name);
      }

      public override bool Equals (object obj)
      {
	return Object.ReferenceEquals (this.name, obj);
      }

      public override int GetHashCode ()
      {
	return name.GetHashCode ();
      }

      public static NameTable Table { get { return nt; } }

      public override string ToString () { return name; }
    }

    private static Symbol Nexpr = "expr";

    private static Symbol Nnull = "";
    private static Symbol Nfuncall = "funcall";
    private static Symbol Nvarref = "varref";
    private static Symbol Ninteger = "integer";
    private static Symbol Nstring = "string";
    private static Symbol Nsymbol = "symbol";
    private static Symbol Nlist = "list";

    private static Symbol Ndefun = "defun";
    private static Symbol Nfname = "fname";
    private static Symbol Nargs = "args";
    private static Symbol Nargs_unevalled = "args-unevalled";
    private static Symbol Nfixed = "fixed";
    private static Symbol Noptional = "optional";
    private static Symbol Nrest = "rest";

    private static Symbol Ndefvar = "defvar";
    private static Symbol Nvname = "vname";
    private static Symbol Ndescription = "description";
    private static Symbol Nrange = "range";

    public abstract class Function
    {
      public Symbol name;
      public int min_arg, max_arg;

      public Function () { }

      public Function (Symbol name, int min_arg, int max_arg)
      {
	this.name = name;
	this.min_arg = min_arg;
	this.max_arg = max_arg;
      }

      public abstract Term Call (Domain domain, Variable vari, Term[] args);

      public override string ToString ()
      {
	return name;
      }

      internal class Subroutine : Function
      {
	public Builtin builtin;
	public bool setvar;

	public Subroutine (Builtin builtin, Symbol name, bool setvar,
			   int min_arg, int max_arg)
	  : base (name, min_arg, max_arg)
	  {
	    this.builtin = builtin;
	  }

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  args = (Term[]) args.Clone ();
	  for (int i = 0; i < args.Length; i++)
	    {
	      args[i] = args[i].Eval (domain);
	      if (domain.Thrown)
		return args[i];
	    }
	  return builtin (domain, vari, args);
	}
      }

      internal class SpecialForm : Function
      {
	public Builtin builtin;

	public SpecialForm (Builtin builtin, Symbol name,
			    int min_arg, int max_arg)
	  : base (name, min_arg, max_arg)
	  {
	    this.builtin = builtin;
	  }

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  return builtin (domain, vari, args);
	}
      }

      internal class Lambda : Function
      {
	internal bool args_evalled;
	internal Variable[] args;
	internal Term[] body;

	public Lambda (Domain domain, XmlNode node)
	  {
	    int nfixed = 0;
	    int noptional = 0;
	    int nrest = 0;
	    name = node.Attributes[Nfname].Value;
	    
	    node = node.FirstChild;
	    if (node != null
		&& (node.Name == Nargs || node.Name == Nargs_unevalled))
	      {
		XmlNode n;
		args_evalled = node.Name == Nargs;
		for (n = node.FirstChild; n != null; n = n.NextSibling)
		  {
		    if (n.Name == Nfixed)
		      nfixed++;
		    else if (n.Name == Noptional)
		      noptional++;
		    else if (n.Name == Nrest)
		      nrest++;
		    else
		      throw new Exception ("Invalid argument type: " + n);
		  }
		min_arg = nfixed;
		max_arg = nfixed + noptional + nrest;
		args = new Variable[max_arg];
		n = node.FirstChild;
		for (int i = 0; i < max_arg; n = n.NextSibling)
		  args[i++] = domain.Defvar ((Symbol) n.Attributes[0].Value);
		if (nrest == 1)
		  max_arg = - max_arg;
	      }
	    else
	      {
		min_arg = max_arg = 0;
	      }
	  }

	public Lambda (Domain domain, Symbol name, bool args_evalled, Symbol[] args)
	  {
	    int nfixed = 0;
	    int noptional = 0;
	    int nrest = 0;

	    this.name = name;
	    this.args_evalled = args_evalled;
	    if (args != null)
	      {
		int i = 0;
		for (i = 0; i < args.Length; i++, nfixed++)
		  if (args[i] == Noptional || args[i] == Nrest)
		    break;
		if (i < args.Length)
		  {
		    if (args[i] == Noptional)
		      {
			for (i++; i < args.Length; i++, noptional++)
			  if (args[i] == Nrest)
			    break;
			if (i < args.Length)
			  nrest = 1;
		      }
		  }
		min_arg = nfixed;
		max_arg = nfixed + noptional + nrest;
		this.args = new Variable[max_arg];
		int j;
		for (i = j = 0; j < this.args.Length; i++)
		  if (args[i] != Noptional || args[i] != Nrest)
		    this.args[j++] = domain.Defvar (args[i]);
	      }
	    else
	      {
		min_arg = max_arg = 0;
	      }
	  }

	public void SetBody (Domain domain, XmlNode node)
	{
	  for (node = node.FirstChild; node != null; node = node.NextSibling)
	    if (node.Name != Ndescription
		&& node.Name != Nargs
		&& node.Name != Nargs_unevalled)
	      break;
	  int nterms = 0;
	  for (XmlNode n = node; n != null; n = n.NextSibling)
	    nterms++;
	  if (nterms > 0)
	    {
	      body = new Term[nterms];
	      for (nterms = 0; node != null; node = node.NextSibling, nterms++)
		body[nterms] = new Term (domain, node);
	    }
	}

	public void SetBody (Term[] body)
	{
	  this.body = body;
	}

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  Bindings current = domain.bindings;
	  Term result = Zero;

	  try {
	    int i;
	    if (args_evalled)
	      {
		Term[] newargs = new Term[args.Length];
		for (i = 0; i < min_arg; i++)
		  newargs[i] = args[i].Eval (domain);
		args = newargs;
	      }
	    for (i = 0; i < min_arg; i++)
	      domain.Bind (this.args[i], args[i]);
	    if (body != null)
	      {
		try {
		  if (args_evalled)
		    domain.Catch (CatchTag.Return);
		  foreach (Term term in body)
		    {
		      result = term.Eval (domain);
		      if (domain.Thrown)
			return result;
		    }
		} finally {
		  if (args_evalled)
		    domain.Uncatch ();
		}
	      }
	  } finally {
	    domain.UnboundTo (current);
	  }
	  return result;
	}
      }
    }

    public class Variable : TermValue
    {
      public readonly Symbol name;
      public string desc;
      internal Term default_val;
      internal Term val;
      object range;

      public Variable (Symbol name, Term value)
      {
	this.name = name;
	val = value;
	default_val = Zero;
      }

      public virtual bool ValueP (Term val) { return true; }

      public override TermValue Clone ()
      {
	Variable v = new Variable (name, val);
	v.desc = desc;
	v.default_val = default_val;
	v.range = range;
	return v;
      }

      public Term Value
      {
	get { return val; }
	set {
	  if (! ValueP (value))
	    throw new Exception ("Invalid value: " + value);
	  val = value;
	}
      }

      public Term DefaultValue
      {
	get { return default_val; }
	set {
	  if (! ValueP (value))
	    throw new Exception ("Invalid value: " + value);
	  default_val = value;
	}
      }

      public virtual object Range
      {
	get { return range; }
	set { range = value; }
      }

      public override Term Eval (Domain domain) { return val; }

      public override string ToString () { return name + "(" + val + ")"; }

      public class Int : Variable
      {
	public int[] range;

	private static bool SubsetP (int[] r1, int[] r2)
	  {
	    if (r2 == null)
	      return true;
	    for (int i = 0; i < r1.Length; i += 2)
	      {
		int j;
		for (j = 0; j < r2.Length; j += 2)
		  if (r2[j] <= r1[i] && r2[j + 1] >= r1[i + 1])
		    break;
		if (j >= r2.Length)
		  return false;
	      }
	    return true;
	  }

	private static bool SubsetP (int val, int[] r)
	  {
	    if (r == null)
	      return true;
	    for (int i = 0; i < r.Length; i += 2)
	      if (r[i] <= val && r[i + 1] >= val)
		return true;
	    return false;
	  }

	public Int (Symbol name, string description, int value, int[] range)
	  : base (name, new Term (value))
	  {
	    if (! SubsetP (value, range))
	      throw new Exception ("Invalid value: " + value);
	    desc = description;
	    default_val = val;
	    this.range = range;
	  }

	public override bool ValueP (Term term)
	{
	  if (! term.IsInt)
	    return false;
	  return SubsetP (term.Intval, range);
	}

	public override object Range {
	  get { return range; }
	  set {
	    int[] r = (int[]) value;
	    if (! SubsetP (r, range)
		|| ! SubsetP (val.Intval, r)
		|| ! SubsetP (default_val.Intval, r))
	      throw new Exception ("Invalid range");
	    range = r;
	  }
	}
      }

      public class Str : Variable
      {
	public string[] range;

	private static bool SubsetP (string[] r1, string[] r2)
	  {
	    if (r2 == null)
	      return true;
	    foreach (string s in r1)
	      if (! SubsetP (s, r2))
		return false;
	    return true;
	  }

	private static bool SubsetP (string str, string[] r)
	  {
	    if (r == null)
	      return true;
	    foreach (string s in r)
	      if (str == s)
		return true;
	    return false;
	}

	public Str (Symbol name, string description, string value, string[] range)
	  : base (name, new Term (value))
	  {
	    if (! SubsetP (value, range))
	      throw new Exception ("Invalid value: " + value);
	    desc = description;
	    default_val = val;
	    this.range = range;
	  }

	public override bool ValueP (Term term)
	{
	  if (! (term.objval is string))
	    return false;
	  return SubsetP (term.Strval, range);
	}

	public override object Range {
	  get { return range; }
	  set {
	    string[] r = (string[]) value;
	    if (! SubsetP (r, range)
		|| ! SubsetP (val.Strval, r)
		|| ! SubsetP (default_val.Strval, r))
	      throw new Exception ("Invalid range");
	    range = r;
	  }
	}
      }

      public class Sym : Variable
      {
	public Symbol[] range;

	private static bool SubsetP (Symbol[] r1, Symbol[] r2)
	  {
	    if (r2 == null)
	      return true;
	    foreach (Symbol n in r1)
	      if (! SubsetP (n, r2))
		return false;
	    return true;
	  }

	private static bool SubsetP (Symbol name, Symbol[] r)
	  {
	    if (r == null)
	      return true;
	    foreach (Symbol n in r)
	      if (name == n)
		return true;
	    return false;
	}

	public Sym (Symbol name, string description, Symbol value, Symbol[] range)
	  : base (name, new Term (value))
	  {
	    if (! SubsetP (value, range))
	      throw new Exception ("Invalid value: " + value);
	    desc = description;
	    default_val = val;
	    this.range = range;
	  }

	public override bool ValueP (Term term)
	{
	  if (! (term.objval is Symbol))
	    return false;
	  return SubsetP (term.Symval, range);
	}

	public override object Range {
	  get { return range; }
	  set {
	    Symbol[] r = (Symbol[]) value;
	    if (! SubsetP (r, range)
		|| ! SubsetP (val.Symval, r)
		|| ! SubsetP (default_val.Symval, r))
	      throw new Exception ("Invalid range");
	    range = r;
	  }
	}
      }
    }

    internal class Bindings
    {
      private Variable vari;
      private Term old_value;
      private Bindings next;
	
      private Bindings (Variable vari)
      {
	this.vari = vari;
	old_value = vari.val;
      }
	
      public static Bindings Bind (Bindings bindings, Variable vari, Term val)
      {
	Bindings b = new Bindings (vari);

	b.vari.Value = val;
	b.next = bindings;
	return b;
      }

      internal Bindings UnboundTo (Bindings boundary)
      {
	for (Bindings b = this; b != boundary; b = b.next)
	  b.vari.val = b.old_value;
	return boundary;
      }

      public override string ToString ()
      {
	string str = "(bindings";
	for (Bindings b = this; b != null; b = b.next)
	  str += " " + vari.name + "=" + b.old_value;
	return str + ")";
      }
    }

#if false
    internal class ThrowException : Exception
    {
      Symbol tag;
      public object value;

      public ThrowException (Symbol tag, object value) : base ()
	{
	  this.tag = tag;
	  this.value = value;
	}
    }
#endif

    internal class CatchTag : IEquatable<CatchTag>
    {
      private object val;

      public CatchTag (Symbol name) { val = name.name; }
      private CatchTag (int i) { val = i; }

      public static CatchTag Return = new CatchTag (0);
      public static CatchTag Break = new CatchTag (1);

      public static bool operator== (CatchTag t1, CatchTag t2)
	{ return t1.val == t2.val; }

      public static bool operator!= (CatchTag t1, CatchTag t2)
	{ return t1.val != t2.val; }

      public bool Equals (CatchTag tag) { return this.val == tag.val; }
      public override bool Equals (object val) { return this.val == val; }

      public override int GetHashCode () { return val.GetHashCode (); }
    }

    public class Domain
    {
      public object context;
      public int depth = 0;

      internal Dictionary<Symbol, TermType> termtypes
	= new Dictionary<Symbol, TermType> ();
      internal Dictionary<Symbol, Function> functions
	= new Dictionary<Symbol, Function> ();
      internal Dictionary<Symbol, Variable> variables
	= new Dictionary<Symbol, Variable> ();
      internal Bindings bindings;
      private Stack<CatchTag> catch_stack = new Stack<CatchTag> ();
      private int catch_count = 0;

      internal Domain () { }

      public Domain (object context) : this (basic, context) { }

      public Domain (Domain parent, object context)
      {
	termtypes = new Dictionary<Symbol, TermType> (parent.termtypes);
	functions = new Dictionary<Symbol, Function> (parent.functions);
	variables = new Dictionary<Symbol, Variable> (parent.variables);
	this.context = context;
      }

      internal void Bind (Variable vari, Term value)
      {
	bindings = Bindings.Bind (bindings, vari, value);
      }

      internal void UnboundTo (Bindings boundary)
      {
	if (bindings != null)
	  bindings = bindings.UnboundTo (boundary);
      }

      internal void Catch (CatchTag tag)
      {
	catch_stack.Push (tag);
	catch_count++;
      }

      internal void Uncatch ()
      {
	catch_stack.Pop ();
	if (catch_count > catch_stack.Count)
	  catch_count--;
      }

      public bool Thrown {
	get { return catch_count < catch_stack.Count; }
      }

      internal void ThrowReturn ()
      {
	foreach (CatchTag tag in catch_stack)
	  {
	    catch_count--;
	    if (tag == CatchTag.Return)
	      break;
	  }
      }

      internal void ThrowBreak ()
      {
	if (catch_stack.Peek () != CatchTag.Break)
	  throw new Exception ("No outer loop to break");
	catch_count--;
      }

      internal void ThrowTag (CatchTag tag)
      {
	foreach (CatchTag elt in catch_stack)
	  {
	    catch_count--;
	    if (elt == tag)
	      break;
	  }
      }

      public void DefTerm (Symbol name, TermParser parser)
      {
	termtypes[name] = new TermType (name, parser);
      }

      public void DefSubr (Builtin builtin, string str, bool setvar,
			   int min_arg, int max_arg, params string[] aliases)
      {
	Symbol name = str;
	Function func = new Function.Subroutine (builtin, name, setvar,
						 min_arg, max_arg);
	functions[name] = func;
	foreach (string a in aliases)
	  functions[(Symbol) a] = func;
      }

      public void DefSpecial (Builtin builtin, string str,
			      int min_arg, int max_arg,
			      params string[] aliases)
      {
	Symbol name = str;
	Function func = new Function.SpecialForm (builtin, name,
						  min_arg, max_arg);
	functions[name] = func;
	foreach (string a in aliases)
	  functions[(Symbol) a] = func;
      }

      public void DefAlias (string alias, string str)
      {
	functions[(Symbol) alias] = functions[(Symbol) str];
      }

      public void Defun (Symbol name, bool args_evalled,
			 Symbol[] args, Term[] body, bool prototype)
      {
	Function func;

	if (prototype || ! functions.TryGetValue (name, out func))
	  {
	    func = new Function.Lambda (this, name, args_evalled, args);
	    functions[name] = func;
	  }
	if (! prototype)
	  ((Function.Lambda) func).SetBody (body);
      }

      public void Defun (XmlNode node, bool prototype)
      {
	Symbol name = node.Attributes[Nfname].Value;
	Function func;

	if (prototype || ! functions.TryGetValue (name, out func))
	  {
	    func = new Function.Lambda (this, node);
	    functions[name] = func;
	  }
	if (! prototype)
	  ((Function.Lambda) func).SetBody (this, node);
      }

      public void Defun (Function func)
      {
	functions[func.name] = func;
      }

      public Variable Defvar (XmlNode node)
      {
	Symbol name = node.Attributes[0].Value;
	String desc;
	Variable vari;

	node = node.FirstChild;
	if (node != null && node.Name == Ndescription)
	  {
	    desc = node.InnerText;
	    node = node.NextSibling;
	  }
	else
	  desc = null;
	if (node != null)
	  {
	    Symbol type = node.Name;
	    XmlNodeList range_list = null;
	    int nranges = 0;
	    string val = node.InnerText;

	    node = node.NextSibling;
	    if (node != null)
	      {
		range_list = node.ChildNodes;
		nranges = range_list.Count;
	      }

	    if (type == Ninteger)
	      {
		int intval = parse_integer (val);
		int[] range = null;
		if (range_list != null)
		  {
		    range = new int[nranges * 2];
		    for (int i = 0; i < nranges; i++)
		      {
			XmlNode n = range_list[i];
			if (n.Name == Nrange)
			  {
			    range[i * 2]
			      = parse_integer (n.FirstChild.InnerText);
			    range[i * 2 + 1]
			      = parse_integer (n.LastChild.InnerText);
			  }
			else
			  {
			    range[i * 2]
			      = range[i * 2 + 1]
			      = parse_integer (n.FirstChild.InnerText);
			  }
		      }
		  }

		if (variables.TryGetValue (name, out vari))
		  {
		    if (! (vari is Variable.Int))
		      throw new Exception ("Inalid value");
		    vari = (Variable) vari.Clone ();
		    Term v = new Term (intval);
		    vari.Value = v;
		    vari.DefaultValue = v;
		    if (range != null)
		      vari.Range = range;
		  }
		else
		  vari = new Variable.Int (name, desc, intval, range);
	      }
	    else if (type == Nstring)
	      {
		string[] range = null;
		if (range_list != null)
		  {
		    range = new string[nranges];
		    for (int i = 0; i < nranges; i++)
		      range[i] = range_list[i].FirstChild.InnerText;
		  }

		if (variables.TryGetValue (name, out vari))
		  {
		    if (! (vari is Variable.Str))
		      throw new Exception ("Invalid value");
		    vari = (Variable) vari.Clone ();
		    Term v = new Term (val);
		    vari.Value = v;
		    vari.DefaultValue = v;
		    if (range != null)
		      vari.Range = range;
		  }
		else
		  vari = new Variable.Str (name, desc, val, range);
	      }
	    else if (type == Nsymbol)
	      {
		Symbol[] range = null;
		if (range_list != null)
		  {
		    range = new Symbol[nranges];
		    for (int i = 0; i < nranges; i++)
		      range[i] = range_list[i].FirstChild.InnerText;
		  }

		if (variables.TryGetValue (name, out vari))
		  {
		    if (! (vari is Variable.Sym))
		      throw new Exception ("Invalid value");
		    vari = (Variable) vari.Clone ();
		    Term v = new Term (val);
		    vari.Value = v;
		    vari.DefaultValue = v;
		    if (range != null)
		      vari.Range = range;
		  }
		else
		  vari = new Variable.Sym (name, desc, val, range);
	      }
	    else
	      throw new Exception ("Unknown type: " + type);
	  }
	else
	  {
	    if (variables.TryGetValue (name, out vari))
	      vari = (Variable) vari.Clone ();
	    else
	      vari = new Variable (name, Zero);
	  }
	variables[name] = vari;
	return vari;
      }

      public Variable Defvar (Variable vari)
      {
	variables[vari.name] = vari;
	return vari;
      }

      internal Variable Defvar (Symbol name)
      {
	Variable vari = new Variable (name, Zero);
	variables[name] = vari;
	return vari;
      }

      internal Function GetFunc (Symbol name)
      {
	Function func;

	if (! functions.TryGetValue (name, out func))
	  throw new Exception ("Unknown function: " + name);
	return func;
      }

      public bool CopyFunc (Domain domain, Symbol name)
      {
	Function func = GetFunc (name);

	domain.functions[name] = func;
	return true;
      }

      public void CopyFunc (Domain domain)
      {
	foreach (KeyValuePair<Symbol, Function> kv in functions)
	  domain.functions[kv.Key] = kv.Value;
      }

      public Variable GetVar (Symbol name, bool create)
      {
	Variable vari;

	if (! variables.TryGetValue (name, out vari))
	  {
	    if (! create)
	      return null;
	    variables[name] = vari = new Variable (name, Zero);
	  }
	return vari;
      }

      public override string ToString ()
      {
	string str = "<(functions";
	foreach (KeyValuePair<Symbol, Function> kv in functions)
	  str += " " + kv.Key;
	str += ") (variabls";
	foreach (KeyValuePair<Symbol, Variable> kv in variables)
	  str += " " + kv.Key;
	str += ")";
	if (bindings != null)
	  str += " " + bindings;
	if (context != null)
	  str += " (" + context + ")";
	str += ">";
	return str;
      }

      internal void DebugWrite (bool head, string fmt, params string[] arg)
      {
	if (debug_level > depth)
	  {
	    if (head)
	      {
		Console.WriteLine ();
		for (int i = 0; i < depth; i++)
		  Console.Write (" ");
	      }
	    Console.Write (fmt, arg);
	  }
      }

      public object SaveValues ()
      {
	Dictionary<Variable,Term> values = new Dictionary<Variable,Term> ();

	foreach (KeyValuePair<Symbol,Variable> kv in variables)
	  values[kv.Value] = kv.Value.val.Clone ();
	return values;
      }

      public void RestoreValues (object values)
      {
	foreach (KeyValuePair<Variable,Term> kv
		 in (Dictionary<Variable,Term>) values)
	  kv.Key.val = kv.Value;
      }
    }

    public delegate Term Builtin (Domain domain, Variable vari, Term[] args);

    private static Domain basic = new Domain ();

    static Xexpression ()
    {
      basic.DefTerm ("funcall", Funcall.parser);

      basic.DefSubr (Fset, "set", true, 1, 1, "=");
      basic.DefSubr (Fnot, "not", false, 1, 1, "!");
      basic.DefSubr (Fadd, "add", true, 1, -1, "+");
      basic.DefSubr (Fmul, "mul", true, 1, -1, "*");
      basic.DefSubr (Fsub, "sub", true, 1, -1, "-");
      basic.DefSubr (Fdiv, "div", true, 1, -1, "/");
      basic.DefSubr (Fmod, "mod", true, 1, 2, "%");
      basic.DefSubr (Flogior, "logior", true, 1, -1, "|");
      basic.DefSubr (Flogand, "logand", true, 1, -1, "&");
      basic.DefSubr (Flsh, "lsh", true, 1, 2, "<<");
      basic.DefSubr (Frsh, "rsh", true, 1, 2, ">>");
      basic.DefSubr (Feq, "eq", false, 2, -1, "==");
      basic.DefSubr (Fnoteq, "noteq", false, 2, 2, "!=");
      basic.DefSubr (Flt, "lt", false, 2, -1, "<");
      basic.DefSubr (Fle, "le", false, 2, -1, "<=");
      basic.DefSubr (Fgt, "gt", false, 2, -1, ">");
      basic.DefSubr (Fge, "ge", false, 2, -1, ">=");
      basic.DefSubr (Fappend, "append", true, 0, -1);
      basic.DefSubr (Fconcat, "concat", true, 0, -1);
      basic.DefSubr (Fnth, "nth", false, 2, 2);
      basic.DefSubr (Fcopy, "copy", false, 1, 1);
      basic.DefSubr (Fins, "ins", true, 2, 2);
      basic.DefSubr (Fdel, "del", true, 2, 2);
      basic.DefSubr (Feval, "eval", false, 1, 1);
      basic.DefSubr (Fbreak, "break", false, 0, 1);
      basic.DefSubr (Freturn, "return", false, 0, 1);

      basic.DefSpecial (Fand, "and", 1, -1, "&&");
      basic.DefSpecial (For, "or", 1, -1, "||");
      basic.DefSpecial (Fprogn, "progn", 0, -1, "expr");
      basic.DefSpecial (Fif, "if", 2, 3);
      basic.DefSpecial (Fwhen, "when", 1, -1);
      basic.DefSpecial (Floop, "loop", 1, -1);
      basic.DefSpecial (Fwhile, "while", 1, -1);
      basic.DefSpecial (Fcond, "cond", 1, -1);
      basic.DefSpecial (Fforeach, "foreach", 2, -1);
      basic.DefSpecial (Fquote, "quote", 1, 1);
      basic.DefSpecial (Ftype, "type", 1, 1);
      basic.DefSpecial (Fcatch, "catch", 2, 2);
      basic.DefSpecial (Fthrow, "throw", 1, 2);
    }

    private static Term Fset (Domain domain, Variable vari, Term[] args)
    {
      vari.Value = args[0];
      return vari.val;
    }

    private static Term Fnot (Domain domain, Variable vari, Term[] args)
    {
      return args[0].IsTrue ? Zero : One;
    }

    private static Term Fadd (Domain domain, Variable vari, Term[] args)
    {
      int n = vari == null ? 0 : vari.val.Intval;

      foreach (Term arg in args)
	n += arg.Intval;
      if (vari == null)
	return new Term (n);
      vari.val.intval = n;
      return vari.val;
    }

    private static Term Fmul (Domain domain, Variable vari, Term[] args)
    {
      int n = vari == null ? 1 : vari.val.Intval;
      foreach (Term arg in args)
	n *= arg.Intval;
      if (vari == null)
	return new Term (n);
      vari.val.intval = n;
      return vari.val;
    }

    private static Term Fsub (Domain domain, Variable vari, Term[] args)
    {
      int n, i;

      if (vari == null)
	{
	  n = args[0].Intval;
	  i = 1;
	}
      else
	{
	  n = vari.val.Intval;
	  i = 0;
	}
      while (i < args.Length)
	n -= args[i++].Intval;
      if (vari == null)
	return new Term (n);
      vari.val.intval = n;
      return vari.val;
    }

    private static Term Fdiv (Domain domain, Variable vari, Term[] args)
    {
      int n, i;

      if (vari == null)
	{
	  n = args[0].Intval;
	  i = 1;
	}
      else
	{
	  n = vari.val.Intval;
	  i = 0;
	}
      while (i < args.Length)
	n /= args[i++].Intval;
      if (vari == null)
	return new Term (n);
      vari.val.intval = n;
      return vari.val;
    }

    private static Term Fmod (Domain domain, Variable vari, Term[] args)
    {
      if (vari == null)
	return new Term (args[0].Intval % args[1].Intval);
      vari.val.intval = vari.val.Intval % args[0].Intval;
      return vari.val;
    }

    private static Term Flogior (Domain domain, Variable vari, Term[] args)
    {
      int n = vari == null ? 0 : vari.val.Intval;
      foreach (Term arg in args)
	n |= arg.Intval;
      if (vari == null)
	return new Term (n);
      vari.val.intval = n;
      return vari.val;
    }

    private static Term Flogand (Domain domain, Variable vari, Term[] args)
    {
      int n, i;

      if (vari == null)
	{
	  n = args[0].Intval;
	  i = 1;
	}
      else
	{
	  n = vari.val.Intval;
	  i = 0;
	}
      while (i < args.Length)
	n &= args[i++].Intval;
      if (vari == null)
	return new Term (n);
      vari.val.intval = n;
      return vari.val;
    }

    private static Term Flsh (Domain domain, Variable vari, Term[] args)
    {
      if (vari == null)
	return new Term (args[0].Intval << args[1].Intval);
      vari.val.intval = vari.val.Intval << args[0].Intval;
      return vari.val;
    }

    private static Term Frsh (Domain domain, Variable vari, Term[] args)
    {
      if (vari == null)
	return new Term (args[0].Intval >> args[1].Intval);
      vari.val.intval = vari.val.Intval >> args[0].Intval;
      return vari.val;
    }

    private static Term Feq (Domain domain, Variable vari, Term[] args)
    {
      Term o = args[0];

      if (o.objval == null)
	{
	  for (int i = 1; i < args.Length; i++)
	    if (args[i].objval != null || args[i].intval != o.intval)
	      return Zero;
	}
      else
	{
	  for (int i = 1; i < args.Length; i++)
	    if (o.objval.Equals (args[i].objval))
	      return Zero;
	}
      return One;
    }

    private static Term Fnoteq (Domain domain, Variable vari, Term[] args)
    {
      return Feq (domain, vari, args);
    }

    private static Term Flt (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;

      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n >= n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fle (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;
      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n > n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fgt (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;
      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n <= n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fge (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;
      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n < n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fappend (Domain domain, Variable vari, Term[] args)
    {
      List<Term> list;

      if (vari == null)
	list = new List<Term> ();
      else
	list = vari.val.Listval;

      foreach (Term arg in args)
	{
	  if (arg.IsList)
	    list.AddRange ((List<Term>) arg.objval);
	  else
	    list.Add (arg);
	}
      if (vari == null)
	{
	  Term result;
	  result.intval = 0;
	  result.objval = list;
	  return result;
	}
      return vari.val;
    }

    private static Term Fconcat (Domain domain, Variable vari, Term[] args)
    {
      string str;

      if (vari == null)
	str = "";
      else
	str = vari.val.Strval;

      foreach (Term arg in args)
	{
	  if (arg.IsStr)
	    str += (string) arg.objval;
	  else if (arg.IsList)
	    foreach (Term term in (List<Term>) arg.objval)
	      str += (char) term.Intval;
	  else
	    str += (char) arg.Intval;
	}
      if (vari == null)
	{
	  Term result;
	  result.intval = 0;
	  result.objval = str;
	  return result;
	}
      vari.val.objval = str;
      return vari.val;
    }

    private static Term Fnth (Domain domain, Variable vari, Term[] args)
    {
      Term result;

      if (args[1].IsStr)
	{
	  result.intval = ((string) args[1].objval)[args[0].Intval];
	  result.objval = null;
	}
      else if (args[1].IsList)
	{
	  result = ((List<Term>) args[1].objval)[args[0].Intval];
	}
      else
	throw new Exception ("Term is not enumelable: " + args[1]);
      return result;
    }

    private static Term Fcopy (Domain domain, Variable vari, Term[] args)
    {
      Term result;

      result.intval = 0;
      result.objval = new List<Term> (args[0].Listval);
      return result;
    }

    private static Term Fins (Domain domain, Variable vari, Term[] args)
    {
      if (vari.val.IsStr)
	vari.val.objval
	  = vari.val.Strval.Insert (args[0].Intval, args[1].Strval);
      else if (vari.val.IsList)
	vari.val.Listval.InsertRange (args[0].Intval, args[1].Listval);
      else
	throw new Exception ("term is not collection: " + vari.val);
      return vari.val;
    }

    private static Term Fdel (Domain domain, Variable vari, Term[] args)
    {
      if (vari.val.IsStr)
	vari.val.objval
	  = vari.val.Strval.Remove (args[0].Intval,
				    args[1].Intval - args[0].Intval);

      else if (vari.val.IsList)
	vari.val.Listval.RemoveRange (args[0].Intval,
				      args[1].Intval - args[0].Intval);
      else
	throw new Exception ("term is not collection: " + vari.val);
      return vari.val;
    }

    private static Term Fand (Domain domain, Variable vari, Term[] args)
    {
      foreach (Term arg in args)
	if (! arg.Eval (domain).IsTrue)
	  return Zero;
      return One;
    }

    private static Term For (Domain domain, Variable vari, Term[] args)
    {
      foreach (Term arg in args)
	if (arg.Eval (domain).IsTrue)
	  return One;
      return Zero;
    }

    private static Term Feval (Domain domain, Variable vari, Term[] args)
    {
      return (args[0].Eval (domain));
    }

    private static Term Fprogn (Domain domain, Variable vari, Term[] args)
    {
      Term result = One;

      foreach (Term arg in args)
	result = arg.Eval (domain);
      return result;
    }

    private static Term Fif (Domain domain, Variable vari, Term[] args)
    {
      if (args[0].Eval (domain).IsTrue)
	return args[1].Eval (domain);
      if (args.Length == 2)
	return Zero;
      return args[2].Eval (domain);
    }

    private static Term Fwhen (Domain domain, Variable vari, Term[] args)
    {
      if (! args[0].Eval (domain).IsTrue)
	return Zero;
      Term result = One;
      for (int i = 1; i < args.Length; i++)
	result = args[i].Eval (domain);
      return result;
    }

    private static Term Freturn (Domain domain, Variable vari, Term[] args)
    {
      domain.ThrowReturn ();
      return args.Length == 0 ? Zero : args[0];
    }	  

    private static Term Fbreak (Domain domain, Variable vari, Term[] args)
    {
      domain.ThrowBreak ();
      return args.Length == 0 ? Zero : args[0];
    }	  

    private static Term Floop (Domain domain, Variable vari, Term[] args)
    {
      Term result = Zero;
      try {
	domain.Catch (CatchTag.Break);
	while (! domain.Thrown)
	  foreach (Term arg in args)
	    {
	      result = arg.Eval (domain);
	      if (domain.Thrown)
		return result;
	    }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    private static Term Fwhile (Domain domain, Variable vari, Term[] args)
    {
      Term result = Zero;
      try {
	domain.Catch (CatchTag.Break);
	while (! domain.Thrown && args[0].Eval (domain).IsTrue)
	  for (int i = 1; i < args.Length; i++)
	    {
	      result = args[i].Eval (domain);
	      if (domain.Thrown)
		return result;
	    }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    private static Term Fcond (Domain domain, Variable vari, Term[] args)
    {
      foreach (Term arg in args)
	{
	  List<Term> list = arg.Listval;
	  Term result = list[0].Eval (domain);

	  if (result.IsTrue)
	    {
	      for (int i = 1; i < list.Count; i++)
		result = list[i].Eval (domain);
	      return result;
	    }
	}
      return Zero;
    }

    private static Term Fforeach (Domain domain, Variable vari,
					Term[] args)
    {
      Term result = args[0].Eval (domain);
      if (domain.Thrown)
	return result;
      List<Term> list = result.Listval;
      Bindings current = domain.bindings;

      try {
	domain.Catch (CatchTag.Break);
	foreach (Term term in list)
	  {
	    domain.Bind (vari, term);
	    try {
	      for (int i = 1; i < args.Length; i++)
		{
		  result = args[i].Eval (domain);
		  if (domain.Thrown)
		    return result;
		}
	    } finally {
	      domain.UnboundTo (current);
	    }
	  }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    private static Term Fquote (Domain domain, Variable vari, Term[] args)
    {
      return new Term (args[0]);
    }

    private static Term Ftype (Domain domain, Variable vari, Term[] args)
    {
      if (args[0].IsInt)
	return TermInt;
      if (args[0].IsStr)
	return TermStr;
      if (args[0].IsSymbol)
	return TermSymbol;
      if (args[0].IsList)
	return TermList;
      return TermTerm;
    }

    public static Term Fcatch (Domain domain, Variable vari, Term[] args)
    {
      Term result = Zero;
      try {
	domain.Catch (new CatchTag (args[0].Symval));
	result = args[1].Eval (domain);
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    public static Term Fthrow (Domain domain, Variable vari, Term[] args)
    {
      domain.ThrowTag (new CatchTag (args[0].Symval));
      return (args.Length == 1 ? Zero : args[1]);
    }

    public delegate TermValue TermParser (Domain domain, XmlNode node);

    public class TermType
    {
      public readonly Symbol type;
      internal readonly TermParser parser;

      public TermType (Symbol type, TermParser parser)
      {
	this.type = type;
	this.parser = parser;
      }
    }

    public abstract class TermValue
    {
      public virtual Term Eval (Domain domain) { return new Term (this); }
      public abstract TermValue Clone ();
    }

    private class Funcall : TermValue
    {
      internal Function func;
      internal Variable vari;
      internal Term[] args;

      private Funcall (Function func, Variable vari, Term[] args)
	{
	  this.func = func;
	  this.vari = vari;
	  this.args = args;
	}

      public Funcall (Domain domain, Symbol fname, Term[] args)
	{
	  func = domain.GetFunc (fname);
	  this.args = args;
	}

      public Funcall (Domain domain, Symbol fname, Symbol vname, Term[] args)
	{
	  func = domain.GetFunc (fname);
	  int nargs = args.Length;
	  if (nargs < func.min_arg
	      || (func.max_arg >= 0 && nargs > func.max_arg))
	    throw new Exception ("Invalid number of arguments to: "
				 + fname + " " + nargs);
	  this.args = args;
	  if (vname != Nnull)
	    vari = domain.GetVar (vname, true);
	}

      internal static TermValue parser (Domain domain, XmlNode node)
	{
	  Symbol fname = node.Name;
	  Symbol vname = Nnull;
	  XmlAttribute attr;

	  if (fname == Nfuncall)
	    fname = node.Attributes[Nfname].Value;
	  attr = node.Attributes[Nvname];
	  if (attr != null)
	    vname = attr.Value;

	  XmlNodeList nlist = node.ChildNodes;
	  int nargs = nlist.Count;
	  Term[] args = new Term[nargs];
	  for (int i = 0; i < nargs; i++)
	    args[i] = new Term (domain, nlist[i]);
	  return new Funcall (domain, fname, vname, args);
	}

      public override Term Eval (Domain domain)
      {
	domain.DebugWrite (true, "(({0}", func.name);
	for (int i = 0; i < args.Length; i++)
	  domain.DebugWrite (false, " {0}", args[i].ToString ());
	domain.DebugWrite (false, ")");
	domain.depth++;
	Term result = func.Call (domain, vari, args);
	domain.depth--;
	domain.DebugWrite (true, " ==> {0})", result.ToString ());
	return result;
      }

      public override TermValue Clone ()
      {
	return new Funcall (func, vari, args);
      }

      public override string ToString ()
      {
	string str = "<funcall fname=\"" + func.name;
	if (args == null)
	  return str + "\"/>";
	str += "\">";
	foreach (Term e in args)
	  str += e;
	return (str + "</funcall>");
      }
    }

    public struct Term
    {
      public int intval;
      public object objval;

      public Term (int i) { intval = i; objval = null; }
      public Term (Symbol name) { intval = 0; objval = name; }
      public Term (string str) { intval = 0; objval = str; }
      public Term (List<Term> list) { intval = 0; objval = list; }
      public Term (Term term) { intval = term.intval; objval = term.objval; }
      public Term (TermValue obj) { intval = 0; objval = obj; }

      public Term (Domain domain, XmlNode node)
	{
	  Symbol name = node.Name;

	  if (name == Ninteger)
	    {
	      intval = parse_integer (node.InnerText);
	      objval = null;
	    }
	  else
	    {
	      intval = 0;
	      if (name == Nsymbol)
		objval = (Symbol) node.InnerText;
	      else if (name == Nstring)
		objval = node.InnerText.Clone ();
	      else if (name == Nvarref)
		objval = domain.GetVar ((Symbol) node.Attributes[0].Value, true);
	      else if (name == Nlist)
		{
		  List<Term> list = new List<Term> ();
		  for (node = node.FirstChild; node != null;
		       node = node.NextSibling)
		    list.Add (new Term (domain, node));
		  objval = list;
		}
	      else
		{
		  TermType term_type;

		  if (domain.termtypes.TryGetValue (name, out term_type))
		    objval = term_type.parser (domain, node);
		  else
		    objval = Funcall.parser (domain, node);
		}
	    }
	}

      public Term (Domain domain, Symbol fname, Term[] args)
	{
	  intval = 0;
	  objval = new Funcall (domain, fname, args);
	}

      public Term (Domain domain, Symbol fname, Symbol vname, Term[] args)
	{
	  intval = 0;
	  objval = new Funcall (domain, fname, vname, args);
	}

      public object Objval {
	get {
	  if (objval == null)
	    throw new Exception ("term is an integer: " + this);
	  return objval;
	}
      }

      public int Intval {
	get {
	  if (objval != null)
	    throw new Exception ("term is not integer: " + this);
	  return intval;
	}
      }

      public string Strval {
	get {
	  if (! IsStr)
	    throw new Exception ("term is not string: " + this);
	  return (string) objval;
	}
      }

      public string Symval {
	get {
	  if (! IsSymbol)
	    throw new Exception ("term is not symbol: " + this);
	  return (Symbol) objval;
	}
      }

      public List<Term> Listval {
	get {
	  if (! IsList)
	    throw new Exception ("term is not list: " + this);
	  return (List<Term>) objval;
	}
      }

      public bool IsTrue {
	get {
	  return (objval == null
		  ? (intval != 0)
		  : objval is List<Term>
		  ? (((List<Term>) objval).Count != 0)
		  : true);
	}
      }
      public bool IsInt { get { return (objval == null); } }
      public bool IsStr { get { return (objval is string); } }
      public bool IsSymbol { get { return (objval is Symbol); } }
      public bool IsList { get { return (objval is List<Term>); } }

      public bool IsType (Type type)
      {
	return (objval == null ? type == typeof (int)
		: type == objval.GetType ());
      }

      public Term Eval (Domain domain)
      {
	if (objval == null || objval is Symbol || objval is string)
	  return this;
	if (objval is List<Term>)
	  return new Term ((List<Term>) objval);
	return ((TermValue) objval).Eval (domain);
      }

      public Term Clone ()
      {
	if (objval == null || objval is Symbol || objval is string)
	  return this;
	if (objval is List<Term>)
	  {
	    List<Term> list = new List<Term> ();
	    list.InsertRange (0, ((List<Term>) objval));
	    return new Term (list);
	  }
	return new Term (((TermValue) objval).Clone ());
      }

      public override string ToString ()
      {
	string str;

	if (objval == null)
	  str = "<integer>" + intval + "</integer>";
	else if (objval is Symbol)
	  str = "<symbol>" + objval + "</symbol>";
	else if (objval is string)
	  str = "<string>" + objval + "</string>";
	else if (objval is List<Term>)
	  {
	    str = "<list>";
	    foreach (Term e in (List<Term>) objval)
	      str += e;
	    str += "</list>";
	  }
	else if (objval is Funcall)
	  str = "<funcall fname=\"" + ((Funcall) objval).func.name + "\"/>";
	else if (objval is Variable)
	  str = "<variable vname=\"" + ((Variable) objval).name + "\"/>";
	else if (objval is Term)
	  str = "<quote>" + objval + "</quote>";
	else
	  throw new Exception ("invalid Term object: " + objval);
	return str;
      }
    }

    static private Term Zero = new Term (0);
    static private Term One = new Term (1);
    static private Term TermInt = new Term (Ninteger);
    static private Term TermStr = new Term (Nstring);
    static private Term TermSymbol = new Term (Nsymbol);
    static private Term TermList = new Term (Nlist);
    static private Term TermTerm = new Term ((Symbol) "term");

    internal static int parse_integer (string str)
    {
      int len = str.Length;
      bool negative = false;

      if (len <= 1)
	return (len == 0 ? 0 : str[0] - '0');

      int c = str[0];
      int i;

      if (c == '?')
	return str[1];
      if ((c == '0' || c == '#') && str[1] == 'x')
	{
	  i = 0;
	  for (int idx = 2; idx < len; idx++)
	    {
	      c = str[idx];
	      if (c < '0')
		break;
	      else if (c <= '9')
		i = i * 16 + (c - '0');
	      else if (c < 'A')
		break;
	      else if (c <= 'F')
		i = i * 16 + (c - 'A');
	      else if (c < 'a')
		break;
	      else if (c <= 'f')
		i = i * 16 + (c - 'a');
	      else
		break;
	    }
	  return i;
	}
      if (c == '-')
	negative = true;
      i = c - '0';
      for (int idx = 1; idx < len; idx++)
	{
	  c = str[idx];
	  if (c < '0' || c > '9')
	    break;
	  i = i * 10 + (c - '0');
	}
      return negative ? - i : i;
    }

    private Term[] terms;

    private void parse_terms (Domain domain, XmlNode node)
    {
      int nterms = 0;
      for (XmlNode n = node; n != null; n = n.NextSibling)
	if (n.NodeType == XmlNodeType.Element)
	  {
	    if (n.Name == Ndefun)
	      domain.Defun (n, true);
	    else if (n.Name == Ndefvar)
	      domain.Defvar (n);
	    else
	      nterms++;
	  }

      terms = new Term[nterms];
      int i = 0;
      for (XmlNode n = node; n != null; n = n.NextSibling)
	if (n.NodeType == XmlNodeType.Element)
	  {
	    if (n.Name == Ndefun)
	      domain.Defun (n, false);
	    else if (n.Name != Ndefvar)
	      terms[i++]= new Term (domain, n);
	  }
    }

    public Xexpression (Domain domain, XmlNode node)
    {
      parse_terms (domain, node);
    }

    public Xexpression (Domain domain, string url)
    {
      XmlDocument doc = new XmlDocument (Symbol.Table);
      XmlNode node;

      using (XmlTextReader reader = new XmlTextReader (url, Symbol.Table))
	{
	  do {
	    reader.Read ();
	  } while (reader.NodeType != XmlNodeType.None
		   && (reader.NodeType != XmlNodeType.Element
		       || reader.Name != Nexpr));
	  if (reader.NodeType == XmlNodeType.None)
	    throw new Exception ("Node <expr> not found");
	  node = doc.ReadNode (reader);
	}
      parse_terms (domain, node.FirstChild);
    }

    public Term Eval (Domain domain)
    {
      Term result = Zero;

      domain.depth = 0;
      try {
	domain.Catch (CatchTag.Return);
	foreach (Term term in terms)
	  {
	    result = term.Eval (domain);
	    if (domain.Thrown)
	      return result;
	  }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    public override string ToString ()
    {
      string str = "";
      for (int i = 0; i < terms.Length; i++)
	str += terms[i];
      return str;
    }
  }
}
